//
// PING.C -- Ping program using ICMP and RAW Sockets
//

#include <stdio.h>
#include <stdlib.h>
#include <winsock.h>
#include <malloc.h>

#include "ping.h"
//#define NUM_OF_PINGS 20000


// Internal Functions
void Ping(LPCSTR pstrHost, FILE*);
void ReportError(LPCSTR pstrFrom);
int  WaitForEchoReply(SOCKET s);
u_short in_cksum(u_short *addr, int len);
// ICMP Echo Request/Reply functions
void CALLBACK SendEchoRequest(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2);
DWORD	RecvEchoReply(SOCKET, LPSOCKADDR_IN, u_char *, UINT*, UINT*);
SOCKET	  s;
struct    sockaddr_in saDest;
unsigned long int count;
unsigned long int NUM_OF_PINGS;
unsigned long int out_of_order;
MMRESULT  timerID;

// main()
void main(int argc, char **argv)
{
    WSADATA wsaData;
    WORD wVersionRequested = MAKEWORD(1,1);
    int nRet;
	MMRESULT accuracy, accuracy2;
	FILE	*ofp;

	// Check arguments
    if (argc != 4)
    {
		fprintf(stderr,"\nUsage: ping hostname outputfile NUM_OF_PINGS\n");
		return;
    }

	// Init WinSock
    nRet = WSAStartup(wVersionRequested, &wsaData);
    if (nRet)
    {
		fprintf(stderr,"\nError initializing WinSock\n");
		return;
    }

	// Check version
	if (wsaData.wVersion != wVersionRequested)
	{
		fprintf(stderr,"\nWinSock version not supported\n");
		return;
	}

	// Open output file
	ofp = fopen(argv[2], "w");
	if(!ofp)
	{
		fprintf(stderr,"\nError opening outputfile\n");
		return;
	}
	
	printf("\n\n");
	accuracy = timeBeginPeriod(1);
	if(accuracy == TIMERR_NOERROR)
		printf("Success\n");
	else
		printf("Failure\n");
	// Go do the ping
	count=0;
	out_of_order=0;
	NUM_OF_PINGS = atoi(argv[3]);
	Ping(argv[1], ofp);

	accuracy2 = timeEndPeriod(1);
	if(accuracy2 == TIMERR_NOCANDO)
		printf("\nFailure");
	else
		printf("\nSuccess");

	// Close output file
	fclose(ofp);

	// Free WinSock
    WSACleanup();
}


// Ping()
// Calls SendEchoRequest() and
// RecvEchoReply() and prints results
void Ping(LPCSTR pstrHost, FILE* ofp)
{
	LPHOSTENT lpHost;
	struct    sockaddr_in saSrc;
	DWORD	  dwTimeSent;
	DWORD	  dwElapsed;
	DWORD	  dwTimeReceived;
	u_char    cTTL;
	UINT	  Seq;
	UINT	  ID;
	DWORD     nLoop;
	DWORD	  nPacketNum=0;
	int       nRet;
//	DWORD*	  Sender;
//	DWORD*	  Receiver;
	u_short*  RTT;
	const int INTERDEPARTURE = 10;

	//printf("In Ping");
	// Create a Raw socket
	s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (s == SOCKET_ERROR) 
	{
		ReportError("socket()");
		return;
	}
	
	// Lookup host
	lpHost = gethostbyname(pstrHost);
	if (lpHost == NULL)
	{
		fprintf(stderr,"\nHost not found: %s\n", pstrHost);
		return;
	}

	RTT= (u_short*) calloc(NUM_OF_PINGS, sizeof(u_short));
	if(RTT == NULL)
	{
		fprintf(stderr,"\nArray allocation failed");
		return;
	}

/*	Sender= (DWORD*) calloc(NUM_OF_PINGS, sizeof(DWORD));
	if(Sender == NULL)
	{
		fprintf(stderr,"\nSender Array allocation failed");
		free(RTT);
		return;
	}

	Receiver= (DWORD*) calloc(NUM_OF_PINGS, sizeof(DWORD));
	if(Receiver == NULL)
	{
		fprintf(stderr,"\nReceiver Array allocation failed");
		free(RTT);
		free(Sender);
		return;
	}
*/


	// Setup destination socket address
	saDest.sin_addr.s_addr = *((u_long FAR *) (lpHost->h_addr));
	saDest.sin_family = AF_INET;
	saDest.sin_port = 0;

	// Tell the user what we're doing
	printf("\nPinging %s [%s] with %d bytes of data:\n",
				pstrHost,
				inet_ntoa(saDest.sin_addr),
				REQ_DATASIZE);

	// Send ICMP echo request
	//SendEchoRequest(0,0,0,0,0);

	timerID = timeSetEvent( INTERDEPARTURE, 0, SendEchoRequest, NULL, TIME_PERIODIC); 



	// Ping multiple times
	for (nLoop = 0; nLoop < NUM_OF_PINGS; nLoop++)
	{
		//SendEchoRequest(s, &saDest);

		// Use select() to wait for data to be received
		nRet = WaitForEchoReply(s);
		if (nRet == SOCKET_ERROR)
		{
			ReportError("select()");
			break;
		}
		if (!nRet)
		{
			printf("\nTimeOut %lu", nLoop);
			break;
		}

		// Receive reply
		dwTimeSent = RecvEchoReply(s, &saSrc, &cTTL, &ID, &Seq);

		dwTimeReceived = timeGetTime();

//		printf("\n%s", inet_ntoa(saDest.sin_addr));
//		printf(" %s", inet_ntoa(saSrc.sin_addr));

		if(saSrc.sin_addr.s_addr != saDest.sin_addr.s_addr)
		{	
//			printf("\nwrong address %lu %s", nLoop, inet_ntoa(saSrc.sin_addr));
			nLoop--;	// don't count this as received
		}
		else
		{
			// Calculate elapsed time
			dwElapsed = dwTimeReceived - dwTimeSent;

			if( (((DWORD)ID*65536)+(DWORD)Seq ) < nPacketNum )
				out_of_order++;
			else
				nPacketNum = ((DWORD)ID*65536)+(DWORD)Seq;

			//Sender[((DWORD)ID*65536)+(DWORD)Seq] = dwTimeSent;
			//Receiver[((DWORD)ID*65536)+(DWORD)Seq] = dwTimeReceived;
			RTT[((DWORD)ID*65536)+(DWORD)Seq] = dwElapsed;
			//RTT[Seq] = dwElapsed;
		}

		if( (count%100000) == 0)
			printf("\nnLoop=%lu", nLoop);

//		printf("\nReply from: %s: bytes=%d time=%ldms TTL=%d Sent=%ld Recv=%ld Seq=%u", 
//              inet_ntoa(saSrc.sin_addr), REQ_DATASIZE, dwElapsed, cTTL, dwTimeSent, dwTimeReceived, Seq);

	}
	printf("\nExited loop");
	nRet = closesocket(s);
	if (nRet == SOCKET_ERROR)
		ReportError("closesocket()");

	for(nLoop = 0; nLoop < NUM_OF_PINGS; nLoop++)
	{	fprintf(ofp, "%d\n",
//		fprintf(ofp, "\n%ld %ld %ld %d",
//			   nLoop,
//			   Sender[nLoop],
//			   Receiver[nLoop],
			   RTT[nLoop]
			   );
	}
	printf("\n%lu Out of Order Packets",out_of_order);
	// Deallocate memory
	free (RTT);
//	free (Sender);
//	free (Receiver);

	//timerID = timeKillEvent(timerID);
}

// SendEchoRequest()
// Fill in echo request header
// and send to destination
void CALLBACK SendEchoRequest(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2) 
{
	static ECHOREQUEST echoReq;
	static u_short nId = 0;
	static u_short nSeq = 0;
	int nRet;
	LPSOCKADDR_IN lpstToAddr;
	lpstToAddr = &saDest;

	// Fill in echo request
	echoReq.icmpHdr.Type		= ICMP_ECHOREQ;
	echoReq.icmpHdr.Code		= 0;
	echoReq.icmpHdr.Checksum	= 0;
	echoReq.icmpHdr.ID			= nId;
	echoReq.icmpHdr.Seq			= nSeq;

	if(nSeq==65535)
	{
		nId++;
		nSeq=0;
		printf("\nnID=%d count=%u",nId, count);
	}
	else
	{
		nSeq++;
	}

	// Fill in some data to send
/*	for (nRet = 0; nRet < REQ_DATASIZE; nRet++)
		echoReq.cData[nRet] = ' '+nRet;
*/
	// Save tick count when sent
//	echoReq.dwTime				= GetTickCount();
	echoReq.dwTime				= timeGetTime();
//	printf("\n%u TimeSent: %u",echoReq.icmpHdr.Seq, echoReq.dwTime);

	// Put data in packet and compute checksum
	echoReq.icmpHdr.Checksum = in_cksum((u_short *)&echoReq, sizeof(ECHOREQUEST));
	// Send the echo request  								  
	nRet = sendto(s,						/* socket */
				 (LPSTR)&echoReq,			/* buffer */
				 sizeof(ECHOREQUEST),
				 0,							/* flags */
				 (LPSOCKADDR)lpstToAddr, /* destination */
				 sizeof(SOCKADDR_IN));   /* address length */

	if (nRet == SOCKET_ERROR)
	{
		printf("count =%ld",count);
		ReportError("sendto()");
	}
	++count;
	if (count>= NUM_OF_PINGS)
	{
		timerID=timeKillEvent(timerID);
		printf("\ntimeKillEvent");
	}
	return; // (nRet);
}

// RecvEchoReply()
// Receive incoming data
// and parse out fields
DWORD RecvEchoReply(SOCKET s, LPSOCKADDR_IN lpsaFrom, u_char *pTTL, UINT *pID, UINT *pSeq) 
{
	ECHOREPLY echoReply;
	int nRet;
	int nAddrLen = sizeof(struct sockaddr_in);

	// Receive the echo reply	
	nRet = recvfrom(s,					// socket
					(LPSTR)&echoReply,	// buffer
					sizeof(ECHOREPLY),	// size of buffer
					0,					// flags
					(LPSOCKADDR)lpsaFrom,	// From address
					&nAddrLen);			// pointer to address len

	// Check return value
	if (nRet == SOCKET_ERROR) 
		ReportError("recvfrom()");

	// return time sent and IP TTL
	*pTTL = echoReply.ipHdr.TTL;
	*pID = echoReply.echoRequest.icmpHdr.ID;
	*pSeq = echoReply.echoRequest.icmpHdr.Seq;
//	printf(" Return %d",echoReply.echoRequest.icmpHdr.Seq);
	return(echoReply.echoRequest.dwTime);   		
}

// What happened?
void ReportError(LPCSTR pWhere)
{
	fprintf(stderr,"\n %s error: %d\n", pWhere,
		WSAGetLastError());
}


// WaitForEchoReply()
// Use select() to determine when
// data is waiting to be read
int WaitForEchoReply(SOCKET s)
{
	struct timeval Timeout;
	fd_set readfds;

	readfds.fd_count = 1;
	readfds.fd_array[0] = s;
	Timeout.tv_sec = 10;
    Timeout.tv_usec = 0;

	return(select(1, &readfds, NULL, NULL, &Timeout));
}


//
// Mike Muuss' in_cksum() function
// and his comments from the original
// ping program
//
// * Author -
// *	Mike Muuss
// *	U. S. Army Ballistic Research Laboratory
// *	December, 1983

/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
u_short in_cksum(u_short *addr, int len)
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 ) {
		u_short	u = 0;

		*(u_char *)(&u) = *(u_char *)w ;
		sum += u;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}
